--[[
Plugin for showing health frames
]]

----------------------------------
-- Create plugin

local RW = LibStub("AceAddon-3.0"):GetAddon("Raid Watch")
local plugin = RW:Plugin("HealthFrame")
if not plugin then return end

---------------------------------
-- Locals

local LSM = LibStub("LibSharedMedia-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("RW-2.0-Plugins")
local Party
local db


local defaults = {
	profile = {
		FadeInTime = 0.7,
		Width = 280, 
		Height = 46, 
		Spacing = 0,
		Fill = false,
		GrowUp = true,
		UseGlobal = true,
		UseModel = true,
	}
}

for i, setting in pairs(RW.defaults.profile) do
	defaults.profile[i] = setting
end

local inset = defaults.profile.BorderSize > 2 and defaults.profile.Insets or 0
local backdrop = {
	tile = false,
	insets = { left = inset, right = inset, top = inset, bottom = inset },
}

-- UNTESTED
local function rightClick(self, button)
	if button == "RightButton" then
		Plugin:RemoveBar(self)
	end
end

------------------------------------
-- Init

function plugin:OnRegister()
	self.db = RW.db:RegisterNamespace("HealthFrame", defaults)
	db = self.db.profile
end

function plugin:PluginEnable()
	self:RegisterMessage("RW_ConfigMode")
	self:RegisterMessage("RW_DefaultsChanged")
	self:RegisterMessage("RW_AddHealthFrame")
	self:RegisterMessage("RW_RemoveHealthFrame")
	self:RegisterMessage("RW_RemoveAllHealthFrames")
	Party = RW:GetPlugin("Party")
	
	self.runningBars = {}
	self.freeBarObjects = {}
end

function plugin:PluginDisable()
	wipe(self.runningBars)
	wipe(self.freeBarObjects)
end

-------------------------------------
-- Anchor

local function dragStart(self)
	self:StartMoving()
end
local function dragStop(self)
	self:StopMovingOrSizing()
	local s = self:GetEffectiveScale()
	plugin.db.profile.x = self:GetLeft() * s
	plugin.db.profile.y = self:GetBottom() * s
end

local function createAnchor()
	if plugin.anchor then return plugin.anchor end
	local db = plugin.db.profile
	local frame = RW:CreateAnchor(plugin:GetName(), UIParent, "Health frames")
	frame:SetScript("OnDragStart", dragStart)
	frame:SetScript("OnDragStop", dragStop)
	frame:ClearAllPoints()
	local s = frame:GetEffectiveScale()
	if db.x then
		frame:SetPoint("BOTTOMLEFT", UIParent, "BOTTOMLEFT", db.x / s, db.y / s)
	else
		frame:SetPoint("TOPRIGHT", UIParent, "TOPRIGHT", -400, -400)
	end
	frame:SetWidth(db.Width)
	plugin.anchor = frame
	return frame
end

function plugin:Reset()
	db.x = nil
	db.y = nil
	createAnchor():ClearAllPoints()
	createAnchor():SetPoint("TOPRIGHT", UIParent, "TOPRIGHT", -400, -400)	
end

local function cID(uid)
	if not uid then return end
	local guid = UnitGUID(uid)
	if guid then return tonumber(guid:sub(9, 12), 16) end
end

local function onMouseOver(self)
	for i, bar in pairs(plugin.runningBars) do
		if cID("mouseover") == bar.trackID then
			bar.uID = "mouseover"
			if not bar.foundData then plugin:UpdateData(bar) end
			plugin:UpdateBar(bar, UnitHealth("mouseover"), UnitHealthMax("mouseover"))
		end
	end
end

local function onUpdate(self, elapsed)
	if self.FadeInTime then
		self.FadeInTime = self.FadeInTime - elapsed
		self:SetAlpha(1 - self.FadeInTime / plugin.db.profile.FadeInTime)
		if self.FadeInTime < 0 then
			self.FadeInTime = nil
		end
	end
	self.t = (self.t or 0) + elapsed
	if self.t > 0.5 then
		for i, bar in pairs(plugin.runningBars) do
			local id = bar.uID
			id = self.uID
			if type(bar.trackID) ~= "function" then
				if cID(id or "") == bar.trackID then
					plugin:UpdateBar(bar, UnitHealth(id or ""), UnitHealthMax(id or ""))
				else
					local uId = Party:GetPartyType()
					for i = 0, Party:GetPartySize() do
						id = (i == 0 and "target") or uId..i.."target"
						if cID(id) == bar.trackID then
							bar.uID = id
							if not bar.foundData then plugin:UpdateData(bar) end
							plugin:UpdateBar(bar, UnitHealth(id), UnitHealthMax(id))
							break
						end
					end
				end
			else
				plugin:UpdateFromFunc(bar, bar.trackID())
			end
		end
		self.t = 0
	end
end

function plugin:ApplySettings(bar)
	local db = self.db.profile
	
	bar:SetWidth(db.Width)
	bar:SetHeight(db.Height)
	
	local inset
	if db.BorderSize < 2 then
		inset = 0
	else
		inset = db.Insets
	end
	
	backdrop.bgFile = LSM:Fetch("background", db.BGTexture)
	backdrop.edgeFile = LSM:Fetch("border", db.Border)
	backdrop.edgeSize = db.BorderSize
	bar:SetBackdrop(backdrop)
	bar:SetBackdropColor(unpack(db.BGColor))
	bar:SetBackdropBorderColor(unpack(db.BorderColor))	
	
	bar.bar:SetStatusBarTexture(LSM:Fetch("statusbar", db.BarTexture))
	if self.color then
		self:SetColor(self.color)
	else
		bar.bar:SetStatusBarColor(unpack(db.BarColor))
	end
	
	bar.timeText:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)
	bar.text:SetFont(LSM:Fetch("font", db.Font), db.FontSize, db.FontOutline)
	bar.text:SetJustifyH(db.FontAlign)
	bar.text:SetTextColor(unpack(db.FontColor))
	
	if bar.iconTexture then
		bar.icon:SetTexture(bar.iconTexture)
		bar.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)	
	elseif db.UseModel then
		bar.model:Show()
		if bar.foundData then
			bar.model:SetCamera(0)
		else
			bar.model:SetModel([[Interface\Buttons\talktomequestionmark.m2]])
			bar.model:SetModelScale(5)
			bar.model:SetPosition(0, 0, -1.3)
		end
		bar.icon:Hide()
	else
		bar.model:Hide()
		bar.icon:Show()
	end
end

function plugin:SetPosition()
	local db = self.db.profile
	local anchor = createAnchor() 
	
	local margin
	if db.BorderSize < 2 or db.Border == "None" then 
		margin = 0 else margin = db.BorderMargin 
	end
	
	for i, bar in pairs(self.runningBars) do
		bar:ClearAllPoints()
		if db.GrowUp then
			bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0 , db.Height * (i-1) + db.Spacing * (i-1))
		else
			bar:SetPoint("BOTTOMLEFT", anchor, "TOPLEFT", 0, -anchor:GetHeight() - db.Height - db.Height * (i-1) - db.Spacing * (i-1))
		end
		bar:SetWidth(db.Width)
		bar:SetHeight(db.Height)
		bar.bar:SetWidth(bar:GetWidth() - bar:GetHeight() - margin)
	end
	anchor:SetWidth(db.Width)
end

function plugin:RemoveBar(bar)
	bar.name = nil
	bar.iconTexture = nil
	bar.FadeInTime = nil
	bar.trackID = nil
	bar.foundData = nil
	bar.uID = nil
	bar.hType = nil
	bar:Hide()
	bar.text:SetText()
	bar.timeText:SetText()
	bar.icon:Show()
	bar.icon:SetTexture([[Interface\Icons\INV_Misc_QuestionMark]])
	bar.model:Show()
	bar.bar:SetValue(1)
	RW:TRemove(self.runningBars, bar)
	tinsert(self.freeBarObjects, bar)
	self:SetPosition()
end

function plugin:GetBar(trackID)
	for i, bar in pairs(self.runningBars) do
		if bar.trackID == trackID then return bar end
	end
end

function plugin:UpdateData(bar)
	bar.name = UnitName(bar.uID)
	bar.model:Show()
	bar.model:SetUnit(bar.uID)
	bar.model:SetCamera(0)
	if not self.db.profile.UseModel then bar.model:Hide() end
	SetPortraitTexture(bar.icon, bar.uID)
	if bar.name then bar.foundData = true end
end

local function formatSmall(value)
	local result

	if value > 10000000 then
	   result = ("%.fm"):format(value/1000000)
	elseif value > 1000000 then
	   result = ("%.2fm"):format(value/1000000)
	elseif value > 100000 then
	   result = ("%.0fk"):format(value/1000)
	elseif value > 1000 then
	   result = ("%.1fk"):format(value/1000)
	end
	return result or value
end

function plugin:UpdateBar(bar, minHealth, maxHealth)
	local percent = minHealth / maxHealth
	bar.text:SetText((bar.name and bar.name or "")..": "..formatSmall(minHealth).."/"..formatSmall(maxHealth))
	bar.timeText:SetText(percent > 0.001 and (math.floor(100 * percent)).."%" or "0%")
	bar.bar:SetValue(percent)
end

function plugin:UpdateFromFunc(bar, ...)
	if bar.hType == "STACK" then
		local stackCount, maxStack = ...
		bar.text:SetText((bar.name and bar.name or ""))
		bar.timeText:SetText((stackCount or 0).."/"..(maxStack or 0))
		bar.bar:SetValue((stackCount or 1) / (maxStack or 1))
	else
		local percent = ...
		bar.text:SetText((bar.name and bar.name or ""))
		bar.timeText:SetText(percent > 0.001 and (math.floor(percent * 100)).."%" or "0%")
		bar.bar:SetValue(percent)
	end
end

function plugin:RW_AddHealthFrame(message, trackID, name, icon, hType)
	local bar = self:GetBar(trackID)
	if not bar then
		bar = tremove(self.freeBarObjects)
		if not bar then 
			bar = RW:CreateBar(rightClick) 
			bar.model = CreateFrame("PlayerModel", bar:GetName().."Model", bar)
			bar.model:SetPoint("TOPLEFT", bar, "TOPLEFT", self.db.profile.BorderMargin, -self.db.profile.BorderMargin)
			bar.model:SetPoint("BOTTOMRIGHT", bar.bar, "BOTTOMLEFT")
		end
		bar.trackID = trackID
		
		bar:RegisterEvent("UPDATE_MOUSEOVER_UNIT")
		bar:SetScript("OnEvent", onMouseOver)
		bar:SetScript("OnUpdate", onUpdate)
		tinsert(self.runningBars, bar)
	end
	bar.name = name
	bar.iconTexture = icon
	bar.FadeInTime = self.db.profile.FadeInTime
	
	bar.hType = hType
	
	bar.icon:SetTexture([[Interface\Icons\INV_Misc_QuestionMark]])
	bar.icon:SetTexCoord(0.08, 0.92, 0.08, 0.92)
	
	bar:Show()
	
	self:ApplySettings(bar)
	self:SetPosition()
end

function plugin:RW_RemoveHealthFrame(message, trackID)
	local bar = self:GetBar(trackID)
	if bar then self:RemoveBar(bar) end
end

do	
	local bars = {}
	function plugin:RW_RemoveAllHealthFrames(message)
		for i, bar in pairs(self.runningBars) do tinsert(bars, bar) end
		for i, bar in pairs(bars) do self:RemoveBar(bar) end
		wipe(bars)
	end
end
-- Options --

function plugin:RW_DefaultsChanged(message, option, value)
	if self.db.profile[option] then
		self.db.profile[option] = value
	end
	for i, bar in pairs(self.runningBars) do
		self:ApplySettings(bar)
	end
end

function plugin:RW_ConfigMode(message, start)
	if start then
		createAnchor():Show()
		self:RW_AddHealthFrame("RW_AddHealthFrame", 29212)
		self:RW_AddHealthFrame("RW_AddHealthFrame", 29205)
	else
		createAnchor():Hide()
		self:RW_RemoveHealthFrame("RW_RemoveHealthFrame", 29212)
		self:RW_RemoveHealthFrame("RW_RemoveHealthFrame", 29205)
	end
end

function plugin:GetOptions()
	local db = self.db.profile
	self.category = "Alert plugins"
	
	local function get(info)
		return db[info[#info]]
	end
	local function set(info, value)
		db[info[#info]] = value
		plugin:RW_DefaultsChanged()
		plugin:SetPosition()
	end
	
	local options = {
		name = "Health Frame",
		type = "group",
		get = get,
		set = set,
		args = {
			Reset = {
				type = "execute",
				name = "Reset",
				func = function() plugin:Reset() end,
			},
			Size = {
				type = "group",
				name = L.SIZE_SETTINGS,
				guiInline = true,
				order = 3,
				args = {
					Width = {
						type = "range",
						name = L.WIDTH,
						min = 10,
						max = 600,
						step = 1,
						order = count,
						},
					Height = {
						type = "range",
						name = L.HEIGHT,
						min = 10,
						max = 600,
						step = 1,
						order = count,
					}
				}
			},
			UseModel = {
				type = "toggle",
				name = L.USE_MODEL,
				desc = L.USE_MODEL_DESC,
				order = 1,
			},
			GrowUp = {
				type = "toggle",
				name = L.GROW_UP,
				order = 2,
			}
		}
	}
	return options
end
